#!/usr/bin/env python

"""
Annotate simplified AST structures. The code in this module operates upon nodes
which are produced when simplifying AST node trees originating from the compiler
module.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from simplified import *
import compiler

class System:

    "A class maintaining the state of the annotation system."

    def __init__(self):
        self.count = 0
    def init(self, node):
        if not hasattr(node, "types"):
            node.types = []
    def annotate(self, node, types):
        self.init(node)
        for type in types:
            if type not in node.types:
                node.types.append(type)
                self.count += 1

system = System()

# Namespaces and related abstractions.

class Namespace:

    """
    A local namespace which may either relate to a genuine set of function
    locals or the initialisation of a structure.
    """

    def __init__(self, structure=None):
        self.structure = structure
        if structure is not None:
            self.local = "structure"
        else:
            self.local = "local"
        self.names = {}
        self.not_local = []

    def make_global(self, name):
        if name not in self.not_local:
            self.not_local.append(name)

    def find_for_store(self, name):
        if name not in self.not_local:
            return self.local
        else:
            return "global"

    def find_for_load(self, name):
        if name not in self.not_local and self.names.has_key(name):
            return self.local
        else:
            return "global"

    def store(self, name, types):
        if name not in self.not_local:
            self.names[name] = types
        else:
            raise KeyError, name

    def load(self, name):
        if name in self.not_local or not self.names.has_key(name):
            raise KeyError, name
        else:
            return self.names[name]

    def merge(self, namespace):
        self.merge_items(namespace.names.items())

    def merge_items(self, items):
        for name, types in items:
            if not self.names.has_key(name):
                self.names[name] = types
            else:
                existing = self.names[name]
                for type in types:
                    if type not in existing:
                        existing.append(type)

class Attribute:

    """
    An attribute abstraction, indicating the type of the attribute along with
    its context or origin.
    """

    def __init__(self, context, type):
        self.context = context
        self.type = type

    def __eq__(self, other):
        return hasattr(other, "type") and other.type == self.type or other == self.type

# Annotation.

class Annotator(Visitor):

    """
    The type annotator which traverses the program nodes, typically depth-first,
    and maintains a record of the current set of types applying to the currently
    considered operation. Such types are also recorded on the nodes, and a
    special "system" record is maintained to monitor the level of annotation
    activity with a view to recognising when no more annotations are possible.
    """

    def __init__(self):
        Visitor.__init__(self)
        self.system = system
        self.types = None
        self.temp = {}

    def process(self, node, locals=None, globals=None):

        """
        Process a subprogram or module 'node', indicating any initial 'locals'
        and 'globals' if either are defined. Return an annotated subprogram or
        module. Note that this method may mutate nodes in the original program.
        """

        # Obtain a namespace either based on locals or on a structure.

        self.namespace = Namespace(structure=getattr(node, "structure", None))
        if locals is not None:
            self.namespace.merge(locals)

        # Determine the global namespace.

        self.global_namespace = globals or self.namespace # NOTE: Improve this.
        node.namespace = self.namespace

        # Add namespace details to any structure involved.

        if hasattr(node, "structure") and node.structure is not None:
            node.structure.namespace = self.namespace

        # Satisfy visitor issues and then dispatch.

        self.visitor = self
        result = self.dispatch(node)

        return result

    def annotate(self, node):

        "Annotate the given 'node' in the system."

        self.system.annotate(node, self.types)

    # Visitor methods.

    def default(self, node):

        """
        Process the given 'node', given that it does not have a specific
        handler.
        """

        for attr in ("expr", "lvalue", "test", "handler"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatch(value))
        for attr in ("body", "else_", "finally_", "code"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        return node

    def dispatch(self, node, *args):
        return Visitor.dispatch(self, node, *args)

    def visitGlobal(self, global_):
        for name in global_.names:
            self.namespace.make_global(name)
        return global_

    def visitLoadRef(self, loadref):
        self.types = [loadref.ref]
        self.annotate(loadref)
        return loadref

    def visitLoadName(self, loadname):
        scope = self.namespace.find_for_load(loadname.name)
        if scope == "structure":
            result = self.dispatch(LoadAttr(expr=LoadRef(ref=self.namespace.structure), name=loadname.name))
        elif scope == "global":
            result = self.dispatch(LoadGlobal(name=loadname.name))
        else:
            self.types = self.namespace.load(loadname.name)
            result = loadname
            self.annotate(result)
        return result

    def visitStoreName(self, storename):
        scope = self.namespace.find_for_store(storename.name)
        if scope == "structure":
            return self.dispatch(StoreAttr(lvalue=LoadRef(ref=self.namespace.structure), name=storename.name, expr=storename.expr))
        elif scope == "global":
            return self.dispatch(StoreGlobal(name=storename.name, expr=storename.expr))
        else:
            storename.expr = self.dispatch(storename.expr)
            self.namespace.store(storename.name, self.types)
            return storename

    def visitLoadGlobal(self, loadglobal):
        self.types = self.global_namespace.load(loadglobal.name)
        self.annotate(loadglobal)
        return loadglobal

    def visitStoreGlobal(self, storeglobal):
        storeglobal.expr = self.dispatch(storeglobal.expr)

        # NOTE: This may always be a merge operation.

        self.global_namespace.store(storeglobal.name, self.types)
        return storeglobal

    def visitLoadTemp(self, loadtemp):
        index = getattr(loadtemp, "index", None)
        self.types = self.temp[index]
        self.annotate(loadtemp)
        return loadtemp

    def visitStoreTemp(self, storetemp):
        storetemp.expr = self.dispatch(storetemp.expr)
        index = getattr(storetemp, "index", None)
        self.temp[index] = self.types
        return storetemp

    def visitReleaseTemp(self, releasetemp):
        index = getattr(releasetemp, "index", None)
        del self.temp[index]
        return releasetemp

    def visitLoadAttr(self, loadattr):
        loadattr.expr = self.dispatch(loadattr.expr)
        types = []
        for ref in self.types:
            for type in ref.namespace.load(loadattr.name):
                types.append(Attribute(ref, type))
        self.types = types
        self.annotate(loadattr)
        return loadattr

    def visitStoreAttr(self, storeattr):
        storeattr.expr = self.dispatch(storeattr.expr)
        expr = self.types
        storeattr.lvalue = self.dispatch(storeattr.lvalue)
        for ref in self.types:
            ref.namespace.store(storeattr.name, expr)
        return storeattr

    def visitInvoke(self, invoke):
        invoke.expr = self.dispatch(invoke.expr)
        expr = self.types

        # NOTE: Consider initialiser invocation for classes.

        types = []
        args = []

        # Get type information for regular arguments.

        for arg in invoke.args:
            args.append(self.dispatch(arg))
            types.append(self.types)

        # Get type information for star and dstar arguments.

        if invoke.star is not None:
            param, default = invoke.star
            default = self.dispatch(default)
            invoke.star = param, default
            types.append(default.types)

        if invoke.dstar is not None:
            param, default = invoke.dstar
            default = self.dispatch(default)
            invoke.dstar = param, default
            types.append(default.types)

        invoke.args = args
        invoke.types = expr

        # NOTE: Now locate and invoke the subprogram.

        for subprogram in expr:
            items = self.make_items(invoke, subprogram)
            annotator = Annotator()
            annotator.process(subprogram, self.make_namespace(items), self.global_namespace)

            # NOTE: Annotate the node with invocation details.

        return invoke

    # Utility methods.

    def make_items(self, invocation, subprogram):
        # NOTE: Support star and dstar.
        args = invocation.args
        params = subprogram.params
        items = []
        keywords = {}

        # Process the specified arguments.

        for arg in args:
            if isinstance(arg, Keyword):
                keywords[arg.name] = arg.expr
                continue
            elif params:
                param, default = params[0]
                if arg is None:
                    arg = self.dispatch(default)
            else:
                raise TypeError, "Invocation has too many arguments."
            items.append((param, arg.types))
            params = params[1:]

        # Collect the remaining defaults.

        while params:
            param, default = params[0]
            if keywords.has_key(param):
                arg = keywords[param]
            else:
                arg = self.dispatch(default)
            items.append((param, arg.types))
            params = params[1:]

        # Add star and dstar.

        if invocation.star is not None:
            if subprogram.star is not None:
                param, default = subprogram.star
                items.append((param, invocation.star.types))
            else:
                raise TypeError, "Invocation provides unwanted *args."
        elif subprogram.star is not None:
            param, default = subprogram.star
            items.append((param, self.dispatch(default)))

        if invocation.dstar is not None:
            if subprogram.dstar is not None:
                param, default = subprogram.dstar
                items.append((param, invocation.dstar.types))
            else:
                raise TypeError, "Invocation provides unwanted **args."
        elif subprogram.dstar is not None:
            param, default = subprogram.dstar
            items.append((param, self.dispatch(default)))

        return items

    def make_namespace(self, items):
        namespace = Namespace()
        namespace.merge_items(items)
        return namespace

# vim: tabstop=4 expandtab shiftwidth=4
